package com.nantian.mfp.framework.security.service;

import com.nantian.mfp.framework.constant.SafeConstants;
import com.nantian.mfp.framework.context.MfpContextHolder;
import com.nantian.mfp.framework.err.HttpException;
import com.nantian.mfp.framework.security.authention.MfpAuthenticationDetailsSource;
import com.nantian.mfp.framework.security.repository.MfpJdbcTokenRepositoryImpl;
import com.nantian.mfp.framework.utils.RequestUtils;
import com.nantian.mfp.pub.model.MfpRememberMeToken;
import org.springframework.dao.DataAccessException;
import org.springframework.security.authentication.AuthenticationDetailsSource;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.core.userdetails.UserDetailsService;
import org.springframework.security.web.authentication.rememberme.*;
import org.springframework.util.Assert;
import org.springframework.util.Base64Utils;
import org.springframework.util.StringUtils;

import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.security.SecureRandom;
import java.util.Arrays;
import java.util.Date;

/**
 * 记住我自动登录
 * @author gnalaifei
 * @date 2023.02.22
 */
public class MfpRememberMeServices extends AbstractRememberMeServices {

    private PersistentTokenRepository tokenRepository = new InMemoryTokenRepositoryImpl();
    private AuthenticationDetailsSource<HttpServletRequest,?> authenticationDetailsSource = new MfpAuthenticationDetailsSource();

    private SecureRandom random = new SecureRandom();
    public static final int DEFAULT_SERIES_LENGTH = 16;
    public static final int DEFAULT_TOKEN_LENGTH = 16;
    private int seriesLength = 16;
    private int tokenLength = 16;

    private boolean alwaysRemember = false;

    public MfpRememberMeServices(String key,int tokenValiditySeconds, UserDetailsService userDetailsService, PersistentTokenRepository tokenRepository) {
        super(key, userDetailsService);
        super.setTokenValiditySeconds(tokenValiditySeconds);
        setAuthenticationDetailsSource(authenticationDetailsSource);
        this.tokenRepository = tokenRepository;
    }

    @Override
    protected UserDetails processAutoLoginCookie(String[] cookieTokens, HttpServletRequest request, HttpServletResponse response) {

        if (cookieTokens.length != 2) {
            throw new InvalidCookieException("Cookie token did not contain " + 2 +
                    " tokens, but contained '" + Arrays.asList(cookieTokens) + "'");
        }

        String presentedSeries ="";
        String presentedToken = cookieTokens[1];
        String clientType="browser";
        if (RequestUtils.isAjaxRequest(request)) {
            presentedSeries = request.getHeader(SafeConstants.FW_SAFE_DEVICEID);
            clientType="moclient";
        }else{
            presentedSeries = cookieTokens[0];
        }

        MfpRememberMeToken token=(MfpRememberMeToken) tokenRepository.getTokenForSeries(presentedSeries);

        if (token == null) {
            // No series match, so we can't authenticate using this cookie
            throw new HttpException(475,"tokenCheck105","自动登录身份过期,请重新登录");
        }

        // We have a match for this user/series combination
        if (!presentedToken.equals(token.getTokenValue())) {
            tokenRepository.removeUserTokens(token.getUsername());
            throw new HttpException(475,"tokenCheck115","自动登录身份异常,请重新登录");
        }

        if (token.getDate().getTime() + getTokenValiditySeconds()*1000L < System.currentTimeMillis()) {
            throw new HttpException(475,"tokenCheck120","自动登录身份异常,请重新登录");
        }

        // Token also matches, so login is valid. Update the token value, keeping the *same* series number.
        if (logger.isDebugEnabled()) {
            logger.debug("Refreshing persistent login token for user '" + token.getUsername() + "', series '" +
                    token.getSeries() + "'");
        }

        //适应emm统一登录
        if(MfpContextHolder.getProps("server-level")!=null&&MfpContextHolder.getProps("server-level").equals("slave")){
            logger.debug("非主服务器登录，不刷新token");
        }else{
            MfpRememberMeToken newToken = new MfpRememberMeToken(token.getUsername(), token.getSeries(), generateTokenData(), new Date(),clientType);

            try {
                tokenRepository.updateToken(newToken.getSeries(), newToken.getTokenValue(), newToken.getDate());
                addCookie(newToken, request, response);
            } catch (DataAccessException e) {
                logger.error("Failed to update token: ", e);
                throw new RememberMeAuthenticationException("Autologin failed due to data access problem");
            }
        }
        return getUserDetailsService().loadUserByUsername(token.getUsername());
    }

    @Override
    protected void onLoginSuccess(HttpServletRequest request, HttpServletResponse response, Authentication successfulAuthentication) {
        String username = successfulAuthentication.getName();

        logger.info("创建token 用户名:" + username);

        MfpRememberMeToken persistentToken=null;
        if (RequestUtils.isAjaxRequest(request)&&request.getHeader(SafeConstants.FW_SAFE_DEVICEID)!=null) {
            persistentToken= new MfpRememberMeToken(username, request.getHeader(SafeConstants.FW_SAFE_DEVICEID),generateTokenData(), new Date(),"moclient");
            logger.debug("最新token值:"+persistentToken.getTokenValue());
        }else{
            persistentToken= new MfpRememberMeToken(username, generateSeriesData(), generateTokenData(), new Date(),"browser");
        }

        try {
            MfpRememberMeToken exToken=((MfpJdbcTokenRepositoryImpl)tokenRepository).checkUserTokens(username);
            if(exToken!=null&&exToken.getSeries().equals(persistentToken.getSeries())){
                tokenRepository.createNewToken(persistentToken);
            }else{
                //如果同一个用户已经有了另外一台设备，则清楚该设备token
                tokenRepository.removeUserTokens(username);
                tokenRepository.createNewToken(persistentToken);
            }
            addCookie(persistentToken, request, response);
        } catch (DataAccessException e) {
            logger.error("Failed to save persistent token ", e);
        }
    }

    @Override
    public void logout(HttpServletRequest request, HttpServletResponse response, Authentication authentication) {
        super.logout(request, response, authentication);
        if (authentication != null) {
            this.tokenRepository.removeUserTokens(authentication.getName());
        }

    }

    protected String generateSeriesData() {
        byte[] newSeries = new byte[this.seriesLength];
        this.random.nextBytes(newSeries);
        return new String(Base64Utils.encode(newSeries));
    }

    protected String generateTokenData() {
        byte[] newToken = new byte[this.tokenLength];
        this.random.nextBytes(newToken);
        return new String(Base64Utils.encode(newToken));
    }

    private void addCookie(PersistentRememberMeToken token, HttpServletRequest request, HttpServletResponse response) {
        this.setCookie(new String[]{token.getSeries(), token.getTokenValue()}, this.getTokenValiditySeconds(), request, response);
    }

    public void setSeriesLength(int seriesLength) {
        this.seriesLength = seriesLength;
    }

    public void setTokenLength(int tokenLength) {
        this.tokenLength = tokenLength;
    }

    @Override
    public void setTokenValiditySeconds(int tokenValiditySeconds) {
        Assert.isTrue(tokenValiditySeconds > 0, "tokenValiditySeconds must be positive for this implementation");
        super.setTokenValiditySeconds(tokenValiditySeconds);
    }



    private void addCookie(MfpRememberMeToken token, HttpServletRequest request, HttpServletResponse response) {
        if (RequestUtils.isAjaxRequest(request)) {
            setCookie(new String[] {generateSeriesData(), token.getTokenValue()}, getTokenValiditySeconds(), request, response);
        }else{
            setCookie(new String[] {token.getSeries(), token.getTokenValue()}, getTokenValiditySeconds(), request, response);
        }
    }

    /**
     * 重写自动登录参数获取,支持ajax请求及自定义参数
     * @param request
     * @param parameter
     * @return
     */
    @Override
    protected boolean rememberMeRequested(HttpServletRequest request, String parameter) {
        if (this.alwaysRemember) {
            return true;
        } else {
            String paramValue = null;
            if(RequestUtils.isAjaxRequest(request)){
                paramValue = (String) request.getAttribute(parameter);
            }else{
                paramValue = request.getParameter(parameter);
            }
            if (paramValue == null || !paramValue.equalsIgnoreCase("true") && !paramValue.equalsIgnoreCase("on") && !paramValue.equalsIgnoreCase("yes") && !paramValue.equals("1")) {
                this.logger.debug(String.format("Did not send remember-me cookie (principal did not set parameter '%s')", parameter));
                return false;
            } else {
                return true;
            }
        }
    }

    /**
     * 重写rememberToken获取,支持从header中获取
     * @param request
     * @return
     */
    @Override
    protected String extractRememberMeCookie(HttpServletRequest request) {
        Cookie[] cookies = request.getCookies();
        if (RequestUtils.isAjaxRequest(request)) {
            if(!StringUtils.hasText(request.getHeader(SafeConstants.FW_SAFE_TOKEN))){
                logger.debug("ajax请求header获取MFPToken异常,不进行自动登录操作");
                return null;
            }
            return request.getHeader(SafeConstants.FW_SAFE_TOKEN);
        }else{
            if ((cookies == null) || (cookies.length == 0)) {
                return null;
            }
            for (Cookie cookie : cookies) {
                if (super.getCookieName().equals(cookie.getName())) {
                    return cookie.getValue();
                }
            }

            return null;
        }
    }
}